/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * textblock.cpp: 
 *
 * Contact:
 *   Moonlight List (moonlight-list@lists.ximian.com)
 *
 * Copyright 2007 Novell, Inc. (http://www.novell.com)
 *
 * See the LICENSE file included with the distribution for details.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cairo.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include "file-downloader.h"
#include "textblock.h"
#include "control.h"
#include "runtime.h"
#include "color.h"
#include "utils.h"
#include "debug.h"
#include "uri.h"


// Unicode Line Separator (\u2028)
static const char utf8_linebreak[3] = { 0xe2, 0x80, 0xa8 };
#define utf8_linebreak_len 3


//
// Inline
//

Inline::Inline ()
{
	SetObjectType (Type::INLINE);
	font = new TextFontDescription ();
	downloader = NULL;
	autogen = false;
}

Inline::~Inline ()
{
	CleanupDownloader ();
	
	delete font;
}

void
Inline::CleanupDownloader ()
{
	if (downloader) {
		downloader->RemoveHandler (Downloader::CompletedEvent, downloader_complete, this);
		downloader->Abort ();
		downloader->unref ();
		downloader = NULL;
	}
}

void
Inline::SetFontSource (Downloader *downloader)
{
	if (this->downloader == downloader)
		return;
	
	CleanupDownloader ();
	
	if (downloader) {
		this->downloader = downloader;
		downloader->ref ();
		
		downloader->AddHandler (downloader->CompletedEvent, downloader_complete, this);
		if (downloader->Started () || downloader->Completed ()) {
			if (downloader->Completed ())
				DownloaderComplete ();
		} else {
			// This is what actually triggers the download
			downloader->Send ();
		}
	} else {
		ClearValue (Inline::FontFilenameProperty);
		ClearValue (Inline::FontGUIDProperty);
	}
}

void
Inline::SetFontResource (const char *resource)
{
	Application *application = Application::GetCurrent ();
	const char *guid = NULL;
	Surface *surface;
	char *filename;
	size_t len;
	Uri *uri;
	
	CleanupDownloader ();
	
	uri = new Uri ();
	
	if (!application || !uri->Parse (resource) || !(filename = application->GetResourceAsPath (uri))) {
		if ((surface = GetSurface ()) && (downloader = surface->CreateDownloader ())) {
			downloader->Open ("GET", resource, XamlPolicy);
			SetFontSource (downloader);
			downloader->unref ();
		} else {
			ClearValue (Inline::FontFilenameProperty);
			ClearValue (Inline::FontGUIDProperty);
		}
		
		delete uri;
		
		return;
	}
	
	delete uri;
	
	// check if the resource is an obfuscated font
	len = strlen (resource);
	if (len > 6 && !g_ascii_strcasecmp (resource + len - 6, ".odttf"))
		guid = resource;
	
	SetValue (Inline::FontFilenameProperty, Value (filename));
	if (guid != NULL)
		SetValue (Inline::FontGUIDProperty, Value (guid));
	else
		ClearValue (Inline::FontGUIDProperty);
	
	g_free (filename);
}

void
Inline::OnPropertyChanged (PropertyChangedEventArgs *args, MoonError *error)
{
	if (args->GetProperty ()->GetOwnerType () != Type::INLINE) {
		DependencyObject::OnPropertyChanged (args, error);
		return;
	}
	
	if (args->GetId () == Inline::FontFamilyProperty) {
		FontFamily *family = args->GetNewValue() ? args->GetNewValue()->AsFontFamily () : NULL;
		const char *family_name;
		char *resource;
		
		if (family && family->source && (family_name = strchr (family->source, '#'))) {
			// FontFamily can reference a font resource in the form "<resource>#<family name>"
			resource = g_strndup (family->source, family_name - family->source);
			SetFontResource (resource);
			g_free (resource);
		} else {
			ClearValue (Inline::FontFilenameProperty);
			ClearValue (Inline::FontGUIDProperty);
		}
	}
	
	NotifyListenersOfPropertyChange (args, error);
}

void
Inline::OnSubPropertyChanged (DependencyProperty *prop, DependencyObject *obj, PropertyChangedEventArgs *subobj_args)
{
	if (prop && prop->GetId () == Inline::ForegroundProperty) {
		// this isn't exactly what we want, I don't
		// think... but it'll have to do.
		NotifyListenersOfPropertyChange (prop, NULL);
	} else {
		DependencyObject::OnSubPropertyChanged (prop, obj, subobj_args);
	}
}

bool
Inline::Equals (Inline *item)
{
	if (item->GetObjectType () != GetObjectType ())
		return false;
	
	if (*item->GetFontFamily () != *GetFontFamily ())
		return false;
	
	if (item->GetFontSize () != GetFontSize ())
		return false;
	
	if (item->GetFontStyle () != GetFontStyle ())
		return false;
	
	if (item->GetFontWeight () != GetFontWeight ())
		return false;
	
	if (item->GetFontStretch () != GetFontStretch ())
		return false;
	
	if (item->GetFontFilename () != GetFontFilename ())
		return false;
	
	if (item->GetTextDecorations () != GetTextDecorations ())
		return false;
	
	// this isn't really correct - we should be checking
	// the innards of the foreground brushes, but we're
	// guaranteed to never have a false positive here.
	if (item->GetForeground () != GetForeground ())
		return false;
	
	// OK, as best we can tell - they are equal
	return true;
}

bool
Inline::UpdateFontDescription ()
{
	FontFamily *family = GetFontFamily ();
	const char *family_name = NULL;
	bool changed = false;
	
	if (family && family->source) {
		if (!(family_name = strchr (family->source, '#')))
			family_name = family->source;
		else
			family_name++;
	}
	
	if (font->SetFilename (GetFontFilename (), GetFontGUID ()))
		changed = true;
	
	if (font->SetFamily (family_name))
		changed = true;
	
	if (font->SetStyle (GetFontStyle ()))
		changed = true;
	
	if (font->SetWeight (GetFontWeight ()))
		changed = true;
	
	if (font->SetSize (GetFontSize ()))
		changed = true;
	
	if (font->SetStretch (GetFontStretch ()))
		changed = true;
	
	return changed;
}

void
Inline::downloader_complete (EventObject *sender, EventArgs *calldata, gpointer closure)
{
	((Inline *) closure)->DownloaderComplete ();
}

void
Inline::DownloaderComplete ()
{
	const char *path, *guid = NULL;
	InternalDownloader *idl;
	char *filename;
	size_t len;
	Uri *uri;
	
	// get the downloaded file path (enforces a mozilla workaround for files smaller than 64k)
	if (!(filename = downloader->GetDownloadedFilename (NULL)))
		return;
	
	g_free (filename);
	
	if (!(idl = downloader->GetInternalDownloader ()))
		return;
	
	if (!(idl->GetObjectType () == Type::FILEDOWNLOADER))
		return;
	
	uri = downloader->GetUri ();
	path = uri->GetPath ();
	
	len = path ? strlen (path) : 0;
	
	if (len > 6 && !g_ascii_strcasecmp (path + len - 6, ".odttf")) {
		// if the font file is obfuscated, use the basename of the path as the guid
		if (!(guid = strrchr (path, '/')))
			guid = path;
		else
			guid++;
	}
	
	// If the downloaded file was a zip file, this'll get the path to the
	// extracted zip directory, else it will simply be the path to the
	// downloaded file.
	if (!(path = ((FileDownloader *) idl)->GetUnzippedPath ()))
		return;
	
	SetValue (Inline::FontFilenameProperty, Value (path));
	if (guid != NULL)
		SetValue (Inline::FontGUIDProperty, Value (guid));
	else
		ClearValue (Inline::FontGUIDProperty);
}



//
// Run
//
bool
Run::Equals (Inline *item)
{
	const char *text, *itext;
	
	if (!Inline::Equals (item))
		return false;
	
	itext = ((Run *) item)->GetText ();
	text = GetText ();
	
	// compare the text
	if (text && itext && strcmp (text, itext) != 0)
		return false;
	else if ((text && !itext) || (!text && itext))
		return false;
	
	return true;
}

//
// TextBlock
//

TextBlock::TextBlock ()
{
	SetObjectType (Type::TEXTBLOCK);
	
	downloader = NULL;
	
	layout = new TextLayout ();
	
	actual_height = 0.0;
	actual_width = 0.0;
	setvalue = true;
	was_set = false;
	dirty = true;
}

TextBlock::~TextBlock ()
{
	CleanupDownloader ();
	
	delete layout;
}

void
TextBlock::CleanupDownloader ()
{
	if (downloader) {
		downloader->RemoveHandler (Downloader::CompletedEvent, downloader_complete, this);
		downloader->Abort ();
		downloader->unref ();
		downloader = NULL;
	}
}

void
TextBlock::SetFontSource (Downloader *downloader)
{
	if (this->downloader == downloader)
		return;
	
	ClearValue (TextBlock::FontSourceProperty);
	CleanupDownloader ();
	
	if (downloader) {
		this->downloader = downloader;
		downloader->ref ();
		
		downloader->AddHandler (downloader->CompletedEvent, downloader_complete, this);
		if (downloader->Started () || downloader->Completed ()) {
			if (downloader->Completed ())
				DownloaderComplete ();
		} else {
			// This is what actually triggers the download
			downloader->Send ();
		}
	} else {
		ClearValue (TextBlock::FontFilenameProperty);
		ClearValue (TextBlock::FontGUIDProperty);
		UpdateFontDescriptions ();
		
		dirty = true;
		
		UpdateBounds (true);
		Invalidate ();
	}
}

void
TextBlock::SetFontResource (const char *resource)
{
	Application *application = Application::GetCurrent ();
	Downloader *downloader;
	const char *guid = NULL;
	Surface *surface;
	char *filename;
	size_t len;
	Uri *uri;
	
	ClearValue (TextBlock::FontSourceProperty);
	CleanupDownloader ();
	
	uri = new Uri ();
	
	if (!application || !uri->Parse (resource) || !(filename = application->GetResourceAsPath (uri))) {
		if ((surface = GetSurface ()) && (downloader = surface->CreateDownloader ())) {
			downloader->Open ("GET", resource, XamlPolicy);
			SetFontSource (downloader);
			downloader->unref ();
		} else {
			ClearValue (TextBlock::FontFilenameProperty);
			ClearValue (TextBlock::FontGUIDProperty);
			UpdateFontDescriptions ();
		}
		
		delete uri;
		
		return;
	}
	
	delete uri;
	
	// check if the resource is an obfuscated font
	len = strlen (resource);
	if (len > 6 && !g_ascii_strcasecmp (resource + len - 6, ".odttf"))
		guid = resource;
	
	SetValue (TextBlock::FontFilenameProperty, Value (filename));
	g_free (filename);
	
	if (guid != NULL)
		SetValue (TextBlock::FontGUIDProperty, Value (guid));
	else
		ClearValue (TextBlock::FontGUIDProperty);
	
	UpdateFontDescriptions ();
}

void
TextBlock::Render (cairo_t *cr, Region *region, bool path_only)
{
	//layout->SetAvailableWidth (GetActualWidth ());
	
	cairo_save (cr);
	cairo_set_matrix (cr, &absolute_xform);
	Paint (cr);
	cairo_restore (cr);
}

void
TextBlock::GetSizeForBrush (cairo_t *cr, double *width, double *height)
{
	*width = actual_width;
	*height = actual_height;
}

Point
TextBlock::GetTransformOrigin ()
{
	Point *user_xform_origin = GetRenderTransformOrigin ();
	return Point (actual_width * user_xform_origin->x, 
		      actual_height * user_xform_origin->y);
}

Size
TextBlock::ComputeActualSize ()
{
	Thickness padding = *GetPadding ();
	Size result = FrameworkElement::ComputeActualSize ();

	//if (dirty) {
	if (!LayoutInformation::GetLastMeasure (this)) {
		Size constraint = Size (INFINITY, INFINITY).Min (GetWidth (), GetHeight ());

		constraint = constraint.GrowBy (-padding);
		Layout (constraint);
	}
		//}
	
	result = Size (actual_width, actual_height);
	result = result.GrowBy (padding);

	return result;
};

Size
TextBlock::MeasureOverride (Size availableSize)
{
	//const char *text = layout->GetText ();
	Thickness padding = *GetPadding ();
	Size constraint;
	Size desired;
	
	//if (text && (!strcmp (text, "751 items") || !strncmp (text, "Use your mouse wheel to", 23))) {
	//	printf ("\nTextBlock::MeasureOverride(availableSize = { %f, %f })\n", availableSize.width, availableSize.height);
	//	printf ("\tText = \"%s\";\n", text);
	//}
	
	constraint = availableSize.GrowBy (-padding);
	Layout (constraint);
	
	desired = Size (actual_width, actual_height).GrowBy (padding);
	
	SetActualHeight (desired.height);
	SetActualWidth (desired.width);

	desired = desired.Min (availableSize);
	
	//if (text && (!strcmp (text, "751 items") || !strncmp (text, "Use your mouse wheel", 20)))
	//	printf ("\treturn { %f, %f };\n", desired.width, desired.height);
	
	return desired;
}

Size
TextBlock::ArrangeOverride (Size finalSize)
{
	//const char *text = layout->GetText ();
	Thickness padding = *GetPadding ();
	Size constraint;
	Size arranged;
	
	//if (text && (!strcmp (text, "751 items") || !strncmp (text, "Use your mouse wheel to", 23))) {
	//	printf ("\nTextBlock::ArrangeOverride(finalSize = { %f, %f })\n", finalSize.width, finalSize.height);
	//	printf ("\tText = \"%s\";\n", text);
	//}
	
	constraint = finalSize.GrowBy (-padding);
	Layout (constraint);
	
	arranged = Size (actual_width, actual_height).GrowBy (padding);
	arranged = arranged.Max (finalSize);
	
	//if (text && (!strcmp (text, "751 items") || !strncmp (text, "Use your mouse wheel", 20)))
	//	printf ("\treturn { %f, %f };\n", arranged.width, arranged.height);
	
	return arranged;
}

void
TextBlock::UpdateLayoutAttributes ()
{
	InlineCollection *inlines = GetInlines ();
	TextLayoutAttributes *attrs;
	const char *text;
	int length = 0;
	Inline *item;
	List *runs;
	
	InvalidateMeasure ();
	InvalidateArrange ();
	runs = new List ();
	
	if (inlines != NULL) {
		for (int i = 0; i < inlines->GetCount (); i++) {
			item = inlines->GetValueAt (i)->AsInline ();
			item->UpdateFontDescription ();
			
 			switch (item->GetObjectType ()) {
			case Type::RUN:
				text = ((Run *) item)->GetText ();
				
				if (text && text[0]) {
					attrs = new TextLayoutAttributes ((ITextAttributes *) item, length);
					runs->Append (attrs);
					
					length += strlen (text);
				}
				
				break;
			case Type::LINEBREAK:
				attrs = new TextLayoutAttributes ((ITextAttributes *) item, length);
				runs->Append (attrs);
				
				length += utf8_linebreak_len;
				break;
			default:
				break;
			}
		}
		
		if (inlines->GetCount () > 0)
			was_set = true;
	}
	
	layout->SetText (GetText (), length);
	layout->SetTextAttributes (runs);
}

bool
TextBlock::UpdateFontDescriptions ()
{
	InlineCollection *inlines = GetInlines ();
	bool changed = false;
	Inline *item;
	
	if (inlines != NULL) {
		for (int i = 0; i < inlines->GetCount (); i++) {
			item = inlines->GetValueAt (i)->AsInline ();
			if (item->UpdateFontDescription ())
				changed = true;
		}
		
		if (changed)
			layout->ResetState ();
	}

	//ClearValue (TextBlock::ActualWidthProperty);
	//ClearValue (TextBlock::ActualHeightProperty);
	InvalidateMeasure ();
	InvalidateArrange ();
	UpdateBounds (true);
	dirty = true;
	return changed;
}

void
TextBlock::Layout (Size constraint)
{
	//const char *text = layout->GetText ();
	
	if (was_set && !GetValueNoDefault (TextBlock::TextProperty)) {
		// If the Text property had been set once upon a time,
		// but is currently empty, Silverlight seems to set
		// the ActualHeight property to the font height. See
		// bug #405514 for details.
		TextFontDescription *desc = new TextFontDescription ();
		const char *family_name;
		FontFamily *family;
		TextFont *font;
		
		if ((family = GetFontFamily ()) && family->source) {
			if (!(family_name = strchr (family->source, '#')))
				family_name = family->source;
			else
				family_name++;
		} else {
			family_name = NULL;
		}
		
		desc->SetFilename (GetFontFilename (), GetFontGUID ());
		desc->SetStretch (GetFontStretch ());
		desc->SetWeight (GetFontWeight ());
		desc->SetStyle (GetFontStyle ());
		desc->SetSize (GetFontSize ());
		desc->SetFamily (family_name);
		
		font = desc->GetFont ();
		actual_height = font->Height ();
		actual_width = 0.0;
		font->unref ();
		delete desc;
	} else if (!was_set) {
		// If the Text property has never been set, then its
		// extents should both be 0.0. See bug #435798 for
		// details.
		actual_height = 0.0;
		actual_width = 0.0;
	} else {
		layout->SetMaxWidth (constraint.width);
		layout->Layout ();
		
		layout->GetActualExtents (&actual_width, &actual_height);
	}
	
	//if (text && (!strcmp (text, "751 items") || !strncmp (text, "Use your mouse wheel to", 23)))
	//	printf ("\tTextBlock::Layout(constraint = { %f, %f }) => %f, %f\n", constraint.width, constraint.height, actual_width, actual_height);
	
	dirty = false;
}

void
TextBlock::Paint (cairo_t *cr)
{
	Thickness *padding = GetPadding ();
	Point offset (padding->left, padding->top);
	
	cairo_set_matrix (cr, &absolute_xform);
	layout->Render (cr, GetOriginPoint (), offset);
	
	if (moonlight_flags & RUNTIME_INIT_SHOW_TEXTBOXES) {
		cairo_set_source_rgba (cr, 0.0, 1.0, 0.0, 1.0);
		cairo_set_line_width (cr, 1);
		cairo_rectangle (cr, padding->left, padding->top, actual_width, actual_height);
		cairo_stroke (cr);
	}
}

char *
TextBlock::GetTextInternal (InlineCollection *inlines)
{
	const char *text;
	GString *block;
	Inline *item;
	char *str;
	
	if (!inlines)
		return g_strdup ("");
	
	block = g_string_new ("");
	
	for (int i = 0; i < inlines->GetCount (); i++) {
		item = inlines->GetValueAt (i)->AsInline ();
		
		switch (item->GetObjectType ()) {
		case Type::RUN:
			text = ((Run *) item)->GetText ();
			
			if (text && text[0])
				g_string_append (block, text);
			break;
		case Type::LINEBREAK:
			g_string_append_len (block, utf8_linebreak, utf8_linebreak_len);
			break;
		default:
			break;
		}
	}
	
	str = block->str;
	g_string_free (block, false);
	
	return str;
}

bool
TextBlock::SetTextInternal (const char *text)
{
	InlineCollection *inlines;
	Value *value;
	Run *run;
	
	// Note: calling GetValue() may cause the InlineCollection to be
	// autocreated, so we need to prevent reentrancy here.
	setvalue = false;
	
	value = GetValue (TextBlock::InlinesProperty);
	inlines = value->AsInlineCollection ();
	inlines->Clear ();
	
	inlines = value->AsInlineCollection ();
	inlines->Clear ();
	
	if (text) {
		run = new Run ();
		run->SetText (text);
		inlines->Add (run);
	} else {
		// setting text to null results in String.Empty
		SetValue (TextBlock::TextProperty, Value (""));
	}
	
	setvalue = true;
	
	return true;
}

void
TextBlock::OnPropertyChanged (PropertyChangedEventArgs *args, MoonError *error)
{
	//if (dirty)
	//	g_error ("entering changed while dirty");

	bool invalidate = true;
	if (args->GetProperty ()->GetOwnerType () != Type::TEXTBLOCK) {
		FrameworkElement::OnPropertyChanged (args, error);
		/*
		if (args->GetId () == FrameworkElement::WidthProperty) {
			//if (layout->SetMaxWidth (args->GetNewValue()->AsDouble ()))
			//	dirty = true;
			
			UpdateBounds (true);
		}
		*/
		return;
	}

	if (args->GetId () == TextBlock::FontFamilyProperty) {
		FontFamily *family = args->GetNewValue() ? args->GetNewValue()->AsFontFamily () : NULL;
		const char *family_name = NULL;
		char *resource;
		
		if (family && family->source && (family_name = strchr (family->source, '#'))) {
			// FontFamily can reference a font resource in the form "<resource>#<family name>"
			resource = g_strndup (family->source, family_name - family->source);
			SetFontResource (resource);
			g_free (resource);
		}
		
		if (UpdateFontDescriptions ())
			dirty = true;
	} else if (args->GetId () == TextBlock::FontSizeProperty) {
		if (UpdateFontDescriptions ())
			dirty = true;
	} else if (args->GetId () == TextBlock::FontStretchProperty) {
		if (UpdateFontDescriptions ())
			dirty = true;
	} else if (args->GetId () == TextBlock::FontStyleProperty) {
		if (UpdateFontDescriptions ())
			dirty = true;
	} else if (args->GetId () == TextBlock::FontWeightProperty) {
		if (UpdateFontDescriptions ())
			dirty = true;
	} else if (args->GetId () == TextBlock::TextProperty) {
		if (setvalue) {
			// result of a change to the TextBlock.Text property
			const char *text = args->GetNewValue() ? args->GetNewValue()->AsString () : NULL;
			
			if (!SetTextInternal (text)) {
				// no change so nothing to invalidate
				invalidate = false;
			} else {
				UpdateLayoutAttributes ();
				dirty = true;
			}
		} else {
			// result of a change to the TextBlock.Inlines property
			UpdateLayoutAttributes ();
			invalidate = false;
		}
	} else if (args->GetId () == TextBlock::TextDecorationsProperty) {
		dirty = true;
	} else if (args->GetId () == TextBlock::TextWrappingProperty) {
		dirty = layout->SetTextWrapping ((TextWrapping) args->GetNewValue()->AsInt32 ());
	} else if (args->GetId () == TextBlock::InlinesProperty) {
		if (setvalue) {
			// result of a change to the TextBlock.Inlines property
			InlineCollection *inlines = args->GetNewValue() ? args->GetNewValue()->AsInlineCollection () : NULL;
			
			setvalue = false;
			// Note: this will cause UpdateLayoutAttributes() to be called in the TextProperty changed logic above
			SetValue (TextBlock::TextProperty, Value (GetTextInternal (inlines), true));
			setvalue = true;
			
			dirty = true;
		} else {
			// result of a change to the TextBlock.Text property
			invalidate = false;
		}
	} else if (args->GetId () == TextBlock::LineStackingStrategyProperty) {
		dirty = layout->SetLineStackingStrategy ((LineStackingStrategy) args->GetNewValue()->AsInt32 ());
	} else if (args->GetId () == TextBlock::LineHeightProperty) {
		dirty = layout->SetLineHeight (args->GetNewValue()->AsDouble ());
	} else if (args->GetId () == TextBlock::TextAlignmentProperty) {
		dirty = layout->SetTextAlignment ((TextAlignment) args->GetNewValue()->AsInt32 ());
	} else if (args->GetId () == TextBlock::PaddingProperty) {
		dirty = true;
	}
	
	if (invalidate) {
		if (dirty) {
			InvalidateMeasure ();
			InvalidateArrange ();
			UpdateBounds (true);
		}

		Invalidate ();
	}
	
	NotifyListenersOfPropertyChange (args, error);
}

void
TextBlock::OnSubPropertyChanged (DependencyProperty *prop, DependencyObject *obj, PropertyChangedEventArgs *subobj_args)
{
	if (prop && prop->GetId () == TextBlock::ForegroundProperty) {
		Invalidate ();
	} else {
		FrameworkElement::OnSubPropertyChanged (prop, obj, subobj_args);
	}
}

void
TextBlock::OnCollectionChanged (Collection *col, CollectionChangedEventArgs *args)
{
	InlineCollection *inlines = GetInlines ();
	
	if (col != inlines) {
		FrameworkElement::OnCollectionChanged (col, args);
		return;
	}
	
	if (args->GetChangedAction () == CollectionChangedActionClearing)
		return;
	
	if (!setvalue) {
		// changes being handled elsewhere...
		return;
	}
	
	setvalue = false;
	// Note: this will cause UpdateLayoutAttributes() to be called in the TextProperty changed logic above
	SetValue (TextBlock::TextProperty, Value (GetTextInternal (inlines), true));
	setvalue = true;
	
	InvalidateMeasure ();
	InvalidateArrange ();
	UpdateBounds (true);
	Invalidate ();
}

void
TextBlock::OnCollectionItemChanged (Collection *col, DependencyObject *obj, PropertyChangedEventArgs *args)
{
	InlineCollection *inlines = GetInlines ();
	
	if (col != inlines) {
		FrameworkElement::OnCollectionItemChanged (col, obj, args);
		return;
	}
	
	if (args->GetId () != Inline::ForegroundProperty) {
		if (args->GetId () == Run::TextProperty) {
			// update our TextProperty
			setvalue = false;
			// Note: this will cause UpdateLayoutAttributes() to be called in the TextProperty changed logic above
			SetValue (TextBlock::TextProperty, Value (GetTextInternal (inlines), true));
			setvalue = true;
		} else {
			// likely a font property change...
			((Inline *) obj)->UpdateFontDescription ();
		}
		
		// All non-Foreground property changes require
		// recalculating layout which can change the bounds.
		InvalidateMeasure ();
		InvalidateArrange ();
		UpdateBounds (true);
		dirty = true;
	} else {
		// A simple Foreground brush change does not require
		// recalculating layout. Invalidate() and we're done.
	}
	
	Invalidate ();
}

void
TextBlock::downloader_complete (EventObject *sender, EventArgs *calldata, gpointer closure)
{
	((TextBlock *) closure)->DownloaderComplete ();
}

void
TextBlock::DownloaderComplete ()
{
	const char *path, *guid = NULL;
	InternalDownloader *idl;
	char *filename;
	size_t len;
	Uri *uri;
	
	dirty = true;
	InvalidateMeasure ();
	InvalidateArrange ();

	// get the downloaded file path (enforces a mozilla workaround for files smaller than 64k)
	if (!(filename = downloader->GetDownloadedFilename (NULL)))
		return;
	
	g_free (filename);
	
	if (!(idl = downloader->GetInternalDownloader ()))
		return;
	
	if (!(idl->GetObjectType () == Type::FILEDOWNLOADER))
		return;
	
	uri = downloader->GetUri ();
	path = uri->GetPath ();
	
	len = path ? strlen (path) : 0;
	
	if (len > 6 && !g_ascii_strcasecmp (path + len - 6, ".odttf")) {
		// if the font file is obfuscated, use the basename of the path as the guid
		if (!(guid = strrchr (path, '/')))
			guid = path;
		else
			guid++;
	}
	
	// If the downloaded file was a zip file, this'll get the path to the
	// extracted zip directory, else it will simply be the path to the
	// downloaded file.
	if (!(path = ((FileDownloader *) idl)->GetUnzippedPath ()))
		return;
	
	SetValue (TextBlock::FontFilenameProperty, Value (path));
	if (guid != NULL)
		SetValue (TextBlock::FontGUIDProperty, Value (guid));
	else
		ClearValue (TextBlock::FontGUIDProperty);
	
	if (UpdateFontDescriptions ()) {
		dirty = true;
		
		UpdateBounds (true);
		Invalidate ();
	}
}
